"""Provides constants for the prefixes used in various IDs in TheOldReader."""

##############################################################################
# Python imports.
from enum import StrEnum


##############################################################################
class Prefix(StrEnum):
    """TheOldReader ID prefixes."""

    FOLDER = "user/-/label/"
    """A folder."""
    FEED = "feed/"
    """A feed."""
    ARTICLE = "tag:google.com,2005:reader/item/"
    """An article."""


##############################################################################
def id_is_a(item_id: str, prefix: Prefix) -> bool:
    """Does the ID look like it's of a particular type?

    Args:
        item_id: The ID to check.
        prefix: The prefix to test against.

    Returns:
        [`True`][True] if the `item_id` has the provided `prefix`,
        [`False`][False] if not.
    """
    return item_id.startswith(prefix)


##############################################################################
def id_is_a_folder(item_id: str) -> bool:
    """Does the ID look like it's a folder?

    Args:
        item_id: The ID to check.

    Returns:
        [`True`][True] if the ID looks like a folder, [`False`][False] if not.
    """
    return id_is_a(item_id, Prefix.FOLDER)


##############################################################################
def id_is_a_feed(item_id: str) -> bool:
    """Does the ID look like it's a feed?

    Args:
        item_id: The ID to check.

    Returns:
        [`True`][True] if the ID looks like a feed, [`False`][False] if not.
    """
    return id_is_a(item_id, Prefix.FEED)


##############################################################################
def id_is_an_article(item_id: str) -> bool:
    """Does the ID look like it's an article?

    Args:
        item_id: The ID to check.

    Returns:
        [`True`][True] if the ID looks like an article, [`False`][False] if not.
    """
    return id_is_a(item_id, Prefix.ARTICLE)


### prefixes.py ends here



"""Provides code for working with lists of article IDs."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, AsyncIterator

##############################################################################
# Local imports.
from .prefixes import Prefix
from .session import Session
from .states import State
from .types import OldList, RawData


##############################################################################
@dataclass(frozen=True)
class ArticleID:
    """Holds an article ID."""

    article_id: str
    """The article ID."""
    direct_stream_ids: list[str]
    """The direct stream IDs."""
    timestamp: datetime
    """The timestamp associated with the article ID."""

    @property
    def full_id(self) -> str:
        """The full, prefixed, ID for the article."""
        return f"{Prefix.ARTICLE}{self.article_id}"

    @classmethod
    def from_json(cls, data: RawData) -> ArticleID:
        """Load the article ID from JSON data.

        Args:
            data: The data to load the article ID from.

        Returns:
            The article ID.
        """
        return cls(
            article_id=data["id"],
            direct_stream_ids=data["directStreamIds"],
            timestamp=datetime.fromtimestamp(
                int(data["timestampUsec"]) / 1_000_000, timezone.utc
            ),
        )


##############################################################################
class ArticleIDs(OldList[ArticleID]):
    """Loads and holds [article ID][oldas.ArticleID] list."""

    @classmethod
    async def stream(
        cls, session: Session, state: State, **filters: Any
    ) -> AsyncIterator[ArticleID]:
        """Stream [article IDs][oldas.ArticleID].

        Args:
            session: The API session object.
            state: The [`State`][oldas.State] to stream.

        Yields:
            The [article IDs][oldas.ArticleID].
        """
        continuation: str | None = ""
        while True:
            result = await session.get(
                "/stream/items/ids", s=str(state), n=10_000, c=continuation, **filters
            )
            for article_id in (
                ArticleID.from_json(article_id)
                for article_id in result.get("itemRefs", [])
            ):
                yield article_id
            if not (continuation := result.get("continuation")):
                break

    @classmethod
    async def load(cls, session: Session, state: State, **filters: Any) -> ArticleIDs:
        """Load [article IDs][oldas.ArticleID].

        Args:
            session: The API session object.
            state: The [`State`][oldas.State] to stream.
            filters: Any addition filter values.

        Returns:
            The list of matching [article IDs][oldas.ArticleID].
        """
        ids: list[ArticleID] = []
        async for article_id in cls.stream(session, state, **filters):
            ids.append(article_id)
        return cls(ids)

    @classmethod
    async def load_read(cls, session: Session) -> ArticleIDs:
        """Load the list of [IDs][oldas.ArticleID] for all read articles.

        Args:
            session: The API session object.

        Returns:
            The list of read [article IDs][oldas.ArticleID].
        """
        return await cls.load(session, State.READ)

    @classmethod
    async def load_unread(cls, session: Session) -> ArticleIDs:
        """Load the list of [IDs][oldas.ArticleID] for all unread articles.

        Args:
            session: The API session object.

        Returns:
            The list of unread [article IDs][oldas.ArticleID].
        """
        return await cls.load(session, State.READING_LIST, xt=State.READ)

    @property
    def full_ids(self) -> list[str]:
        """The list of article IDs.

        Note:
            This is a list of strings where each string is a full article ID.
        """
        return [article.full_id for article in self]

    async def mark_read(self, session: Session) -> bool:
        """Mark all the articles as read.

        Args:
            session: The API session object.

        Returns:
            [`True`][True] if the request to mark as read worked,
            [`False`][False] if not.
        """
        return await session.add_tag(self.full_ids, State.READ)

    async def mark_unread(self, session: Session) -> bool:
        """Mark all the articles as unread.

        Args:
            session: The API session object.

        Returns:
            [`True`][True] if the request to mark as unread worked,
            [`False`][False] if not.
        """
        return await session.remove_tag(self.full_ids, State.READ)


### article_ids.py ends here



"""TheOldReader API async client library."""

##############################################################################
# Python imports.
from importlib.metadata import version

######################################################################
# Main library information.
__author__ = "Dave Pearson"
__copyright__ = "Copyright 2025, Dave Pearson"
__credits__ = ["Dave Pearson"]
__maintainer__ = "Dave Pearson"
__email__ = "davep@davep.org"
__version__: str = version("oldas")
__licence__ = "MIT"

##############################################################################
# Local imports.
from .article_ids import ArticleID, ArticleIDs
from .articles import Article, Articles
from .exceptions import OldASError, OldASInvalidLogin
from .folders import Folder, Folders
from .prefixes import Prefix, id_is_a_feed, id_is_a_folder, id_is_an_article
from .session import Session
from .states import State
from .subscriptions import Subscription, Subscriptions
from .unread import Count, Counts, Unread

##############################################################################
# Exports.
__all__ = [
    "Article",
    "ArticleID",
    "ArticleIDs",
    "Articles",
    "Count",
    "Counts",
    "Folder",
    "Folders",
    "id_is_a_feed",
    "id_is_a_folder",
    "id_is_an_article",
    "OldASError",
    "OldASInvalidLogin",
    "Prefix",
    "Session",
    "State",
    "Subscription",
    "Subscriptions",
    "Unread",
]

### __init__.py ends here



"""Types used across the library."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    Iterator,
    TypeAlias,
    TypeVar,
    overload,
)

##############################################################################
RawData: TypeAlias = dict[str, Any]
"""Type of raw data that comes from TheOldReader."""

##############################################################################
OldData = TypeVar("OldData")
"""Type used in the OldSequence generic wrapper."""


##############################################################################
class OldList(Generic[OldData]):
    """Base class for classes that wrap a list of data."""

    def __init__(self, data: Iterable[OldData] | None = None) -> None:
        """Initialise the OldList object.

        Args:
            data: The initial data to wrap.
        """
        self._data: list[OldData] = list(data or [])
        """The data we're wrapping around."""

    if TYPE_CHECKING:

        @overload
        def __getitem__(self, index: int) -> OldData: ...
        @overload
        def __getitem__(self, index: slice) -> OldList[OldData]: ...

    def __getitem__(self, index: int | slice) -> OldData | OldList[OldData]:
        return (
            OldList[OldData](self._data[index])
            if isinstance(index, slice)
            else self._data[index]
        )

    def __len__(self) -> int:
        return len(self._data)

    def __length_hint__(self) -> int:
        return len(self)

    def __bool__(self) -> bool:
        return bool(self._data)

    def __contains__(self, data: OldData) -> bool:
        return data in self._data

    def __iter__(self) -> Iterator[OldData]:
        return iter(self._data)

    def __repr__(self) -> str:
        return repr(self._data)


### types.py ends here



"""Provides a class for getting article data."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, AsyncIterator, Iterable, Literal

##############################################################################
# Local imports.
from .folders import Folder
from .prefixes import id_is_a_folder
from .session import Session
from .states import State
from .subscriptions import Subscription
from .types import OldList, RawData

##############################################################################
Direction = Literal["ltr", "rtl"]
"""Possible values for the summary direction."""


##############################################################################
@dataclass(frozen=True)
class Summary:
    """The summary details for an [`Article`][oldas.Article]."""

    direction: Direction
    """The direction for the text in the summary."""
    content: str
    """The content of the summary."""

    @classmethod
    def from_json(cls, data: RawData) -> Summary:
        """Load the summary from JSON data.

        Args:
            data: The data to load the summary from.

        Returns:
            The summary.
        """
        return cls(
            direction=data["direction"],
            content=data["content"],
        )


##############################################################################
@dataclass(frozen=True)
class Origin:
    """The origin details for an [`Article`][oldas.Article]."""

    stream_id: str | None
    """The stream ID for the article's origin."""
    title: str
    """The title of the origin of the article."""
    html_url: str
    """The URL of the HTML of the origin of the article."""

    @classmethod
    def from_json(cls, data: RawData) -> Origin:
        """Load the origin from JSON data.

        Args:
            data: The data to load the origin from.

        Returns:
            The origin data.
        """
        return cls(
            stream_id=data.get("streamId"),
            title=data["title"],
            html_url=data["htmlUrl"],
        )


##############################################################################
@dataclass(frozen=True)
class Alternate:
    """Holds details of an alternate for an [`Article`][oldas.Article]."""

    href: str
    """The URL for the alternate."""
    mime_type: str
    """The MIME type of the alternate."""

    @classmethod
    def from_json(cls, data: RawData) -> Alternate:
        """Load the alternate data from JSON data.

        Args:
            data: The data to load the alternate data from.

        Returns:
            The alternates.
        """
        return cls(
            href=data["href"],
            mime_type=data["type"],
        )


##############################################################################
class Alternates(OldList[Alternate]):
    """Holds a list of [alternates][oldas.articles.Alternate] for an [`Article`][oldas.Article]."""


##############################################################################
@dataclass(frozen=True)
class Article:
    """Holds details about an article."""

    id: str
    """The ID of the article."""
    title: str
    """The title of the article."""
    published: datetime
    """The time when the article was published."""
    updated: datetime
    """The time when the article was updated."""
    author: str
    """The author of the article."""
    summary: Summary
    """The summary of the article."""
    categories: list[State | str]
    """The list of categories associated with this article."""
    origin: Origin
    """The origin of the article."""
    alternate: Alternates
    """Alternates for the article."""

    @property
    def is_read(self) -> bool:
        """Has this article been read?"""
        return State.READ in self.categories

    @property
    def is_unread(self) -> bool:
        """Is the article still unread?"""
        return not self.is_read

    @property
    def is_fresh(self) -> bool:
        """Is the article considered fresh?"""
        return State.FRESH in self.categories

    @property
    def is_stale(self) -> bool:
        """Is the article considered stale?"""
        return not self.is_fresh

    @property
    def is_updated(self) -> bool:
        """Does the article look like it's been updated?"""
        return self.published != self.updated

    @property
    def html_url(self) -> str | None:
        """The best guess at the HTML URL for the article."""
        return next(
            (
                alternate.href
                for alternate in self.alternate
                if alternate.mime_type == "text/html"
            ),
            None,
        )

    async def mark_read(self, session: Session) -> bool:
        """Mark the article as read.

        Args:
            session: The API session object.

        Returns:
            [`True`][True] if the request to mark as read worked,
            [`False`][False] if not.
        """
        return await session.add_tag(self.id, State.READ)

    async def mark_unread(self, session: Session) -> bool:
        """Mark the article as unread.

        Args:
            session: The API session object.

        Returns:
            [`True`][True] if the request to mark as unread worked,
            [`False`][False] if not.
        """
        return await session.remove_tag(self.id, State.READ)

    @staticmethod
    def clean_categories(categories: Iterable[str]) -> list[State | str]:
        """Clean up a collection of categories.

        Args:
            categories: The categories to clean up.

        Returns:
            The cleaned categories.

        The incoming list of categories will simply be a list of strings,
        but each of them may refer to a folder or a [state][oldas.State],
        etc. This method will clean the list, turning relevant values into
        their specific type.

        Note:
            For the moment only values matching a [`State`][oldas.State]
            will be turned into their related type.
        """
        return [
            category if id_is_a_folder(category) else State(category)
            for category in categories
        ]

    @classmethod
    def from_json(cls, data: RawData) -> Article:
        """Load the article from JSON data.

        Args:
            data: The data to load the article from.

        Returns:
            The article.
        """
        return cls(
            id=data["id"],
            title=data["title"],
            published=datetime.fromtimestamp(data["published"], timezone.utc),
            updated=datetime.fromtimestamp(data["updated"], timezone.utc),
            author=data["author"],
            summary=Summary.from_json(data["summary"]),
            categories=cls.clean_categories(data["categories"]),
            origin=Origin.from_json(data["origin"]),
            alternate=Alternates(
                Alternate.from_json(alternate) for alternate in data["alternate"]
            ),
        )


##############################################################################
class Articles(OldList[Article]):
    """Loads and holds a full list of [articles][oldas.Article]."""

    @classmethod
    async def stream(
        cls, session: Session, stream: str | Subscription | Folder = "", **filters: Any
    ) -> AsyncIterator[Article]:
        """Load [articles][oldas.Article] from a given stream.

        Args:
            session: The API session object.
            stream: The stream identifier to load from.
            filters: Any other filters to pass to the API.

        Yields:
            The [articles][oldas.Article] matching the request.
        """
        if isinstance(stream, (Folder, Subscription)):
            stream = stream.id
        continuation: str | None = ""
        while True:
            result = await session.get(
                "/stream/contents", s=stream, c=continuation, **filters
            )
            for article in (
                Article.from_json(article) for article in result.get("items", [])
            ):
                yield article
            if not (continuation := result.get("continuation")):
                break

    @classmethod
    async def stream_new_since(
        cls,
        session: Session,
        since: datetime,
        stream: str | Subscription | Folder = "",
        **filters: Any,
    ) -> AsyncIterator[Article]:
        """Stream all [articles][oldas.Article] newer than a given time.

        Args:
            session: The API session object.
            since: Time from which to load articles.
            stream: The stream identifier to stream from.
            filters: Any other filters to pass to the API.

        Yields:
            The [articles][oldas.Article] matching the request.
        """
        async for article in cls.stream(
            session,
            stream,
            ot=int(since.timestamp()),  # codespell:ignore ot,
            # The continuation of "newer than" filtered items seems to not
            # work unless we order the result; so let's go oldest first...
            r="o",
            **filters,
        ):
            yield article


### articles.py ends here



from asyncio import run
from os import getenv

from oldas import Subscriptions

from .session import Session


async def main() -> None:
    if token := getenv("TOR_TOKEN"):
        session = Session("test", token)
    else:
        session = await Session("test").login(
            getenv("TOR_USER", ""), getenv("TOR_PASSWORD", "")
        )
    for subscription in await Subscriptions.load(session):
        print(f"{subscription.title} {subscription.folder_id}")


if __name__ == "__main__":
    run(main())



"""Provides a class for getting subscription information."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from datetime import datetime, timezone

##############################################################################
# Local imports.
from .folders import Folder, Folders
from .prefixes import Prefix, id_is_a_feed, id_is_a_folder
from .session import Session
from .types import OldList, RawData


##############################################################################
@dataclass(frozen=True)
class Category:
    """Holds details of a category."""

    id: str
    """The ID for the category."""
    label: str
    """The label for the category."""

    @classmethod
    def from_json(cls, data: RawData) -> Category:
        """Load the category from JSON data.

        Args:
            data: The data to load the category from.

        Returns:
            The category.
        """
        return cls(
            id=data["id"],
            label=data["label"],
        )


##############################################################################
class Categories(OldList[Category]):
    """Holds a collection of [categories][oldas.subscriptions.Category]."""

    def __contains__(self, data: Category | Folder | str) -> bool:
        """Check if some data is `in` the categories.

        Args:
            data: The category, folder or string to look for.

        Returns:
            [`True`][True] if the data was found, [`False`][False] if not.
        """
        if isinstance(data, (Category, Folder)):
            data = data.id
        return any(category.id == data for category in self)


##############################################################################
@dataclass(frozen=True)
class Subscription:
    """Holds a subscription."""

    id: str
    """The ID of the subscription."""
    title: str
    """The title of the subscription."""
    sort_id: str
    """The sort ID of the subscription."""
    first_item_time: datetime
    """The time of the first item."""
    url: str
    """The URL of the subscription."""
    html_url: str
    """The HTML URL of the subscription."""
    categories: Categories
    """The categories for the subscription."""

    @classmethod
    def from_json(cls, data: RawData) -> Subscription:
        """Load the subscription from JSON data.

        Args:
            data: The data to load the subscription from.

        Returns:
            The subscription.
        """
        return cls(
            id=data["id"],
            title=data["title"],
            sort_id=data["sortid"],
            first_item_time=datetime.fromtimestamp(
                int(data["firstitemmsec"]) / 1_000, timezone.utc
            ),
            url=data["url"],
            html_url=data["htmlUrl"],
            categories=Categories(
                Category.from_json(category) for category in data["categories"]
            ),
        )

    @property
    def folder_id(self) -> str | None:
        """The ID of the folder that this subscription belongs to, or [`None`][None] if it doesn't.

        Note:
            According to the API documentation it would appear that a
            subscription could be a member of multiple folders. Note that
            this property is the ID of the first folder that could be found
            amongst the
            [categories][oldas.subscriptions.Subscription.categories].
        """
        return next(
            (
                category.id
                for category in self.categories
                if id_is_a_folder(category.id)
            ),
            None,
        )


##############################################################################
@dataclass(frozen=True)
class SubscribeResult:
    """Class that holds the request of adding a subscription."""

    query: str
    """The query that was performed."""
    number_of_results: int
    """The number of requests from the query to add."""
    stream_id: str | None
    """The stream ID if the subscription took place."""
    error: str | None
    """The reason why the subscribe failed, if it did."""

    @classmethod
    def from_json(cls, data: RawData) -> SubscribeResult:
        """Load the subscribe result from JSON data.

        Args:
            data: The data to load the subscribe result from.

        Returns:
            The result of making the subscribe request.
        """
        return cls(
            query=data["query"],
            number_of_results=data["numResults"],
            stream_id=data.get("streamId"),
            error=data.get("error"),
        )

    @property
    def failed(self) -> bool:
        """Did the request to subscribe fail?"""
        return self.number_of_results == 0


##############################################################################
class Subscriptions(OldList[Subscription]):
    """Loads and holds the full list of [subscriptions][oldas.Subscription]."""

    @classmethod
    async def load(cls, session: Session) -> Subscriptions:
        """Load the subscriptions.

        Args:
            session: The API session object.

        Returns:
            A list of subscriptions.
        """
        return cls(
            Subscription.from_json(subscription)
            for subscription in (await session.get("subscription/list"))[
                "subscriptions"
            ]
        )

    @staticmethod
    async def add(session: Session, feed: str) -> SubscribeResult:
        """Add a subscription.

        Args:
            session: The API session object.
            feed: The feed to subscribe to.

        Returns:
            A [`SubscribeResult`][oldas.subscriptions.SubscribeResult].

        Notes:
            The `feed` will normally be the URL of the feed to subscribe to.
        """
        return SubscribeResult.from_json(
            await session.post("subscription/quickadd", quickadd=feed)
        )

    @staticmethod
    def _full_id(subscription: str | Subscription) -> str:
        """Get the full ID for a given subscription.

        Args:
            subscription: The subscription to get the full ID for.

        Returns:
            The full ID for the subscription.
        """
        if isinstance(subscription, Subscription):
            subscription = subscription.id
        return (
            subscription
            if id_is_a_feed(subscription)
            else f"{Prefix.FEED}{subscription}"
        )

    @classmethod
    async def rename(
        cls, session: Session, subscription: str | Subscription, new_name: str
    ) -> bool:
        """Rename a subscription.

        Args:
            session: The API session object.
            subscription: The subscription to rename.
            new_name: The new name for the subscription.

        Returns:
            [`True`][True] if the rename call worked, [`False`][False] if not.

        Note:
            The `subscription` can either be a string that is the ID of a feed, or
            it can be a [`Subscription`][oldas.Subscription] object.
        """
        return await session.post_ok(
            "subscription/edit", ac="edit", s=cls._full_id(subscription), t=new_name
        )

    @classmethod
    async def remove(cls, session: Session, subscription: str | Subscription) -> bool:
        """Remove a subscription.

        Args:
            session: The API session object.
            subscription: The subscription to unsubscribe.

        Returns:
            [`True`][True] if the unsubscribe call worked, [`False`][False] if not.

        Note:
            The `subscription` can either be a string that is the ID of a feed, or
            it can be a [`Subscription`][oldas.Subscription] object.
        """
        return await session.post_ok(
            "subscription/edit", ac="unsubscribe", s=cls._full_id(subscription)
        )

    @classmethod
    async def move(
        cls,
        session: Session,
        subscription: str | Subscription,
        target_folder: str | Folder | None = None,
    ) -> bool:
        """Move a subscription to a different folder.

        Args:
            session: The API session object.
            subscription: The subscription to move.
            target_folder: The folder to move the subscription to.

        Returns:
            [`True`][True] if the move call worked, [`False`][False] if not.

        Note:
            If `target_folder` is omitted, is [`None`][None], or is an empty
            [string][str], the subscription will be moved to the top-level
            default folder.
        """
        if isinstance(target_folder, str):
            target_folder = target_folder.strip()
        operation = (
            {"r": "remove"}
            if not target_folder
            else {"a": Folders.full_id(target_folder)}
        )
        return await session.post_ok(
            "subscription/edit", ac="edit", s=cls._full_id(subscription), **operation
        )


### subscriptions.py ends here



"""Provides exception classes for the library."""


##############################################################################
class OldASError(Exception):
    """Base exception for exceptions raised by the library."""


##############################################################################
class OldASInvalidLogin(OldASError):
    """Exception thrown if the login failed."""


##############################################################################
class OldASLoginNeeded(OldASError):
    """Exception thrown when a call is made when a login is needed.

    Note:
        This exception might indicate that either no login has happened yet,
        or it might indicate that a previous login has gone stale and a
        fresh one should take place.
    """


### exceptions.py ends here



"""Provides constants for the states."""

##############################################################################
# Python imports.
from enum import StrEnum


##############################################################################
class State(StrEnum):
    """TheOldReader state names."""

    READ = "user/-/state/com.google/read"
    """An article that has been read."""
    STARRED = "user/-/state/com.google/starred"
    """An article that has been starred."""
    FRESH = "user/-/state/com.google/fresh"
    """An article that is considered fresh."""
    KEPT_UNREAD = "user/-/state/com.google/kept-unread"
    """An article was marked by the user to be kept unread."""
    READING_LIST = "user/-/state/com.google/reading-list"
    """The article is part of the user's reading list."""
    BROADCAST = "user/-/state/com.google/broadcast"
    """The article has been shared."""
    LIKE = "user/-/state/com.google/like"
    """The article has been liked."""


### states.py ends here



"""Provides a class for getting and managing a login session."""

##############################################################################
from typing import Any, Awaitable, Final, Literal, Self

##############################################################################
# Httpx imports.
import httpx
from httpx import AsyncClient, HTTPStatusError, ReadTimeout, RequestError, Response

##############################################################################
# Local imports.
from . import __version__
from .exceptions import OldASError, OldASInvalidLogin, OldASLoginNeeded
from .states import State
from .types import RawData


##############################################################################
class Session:
    """Class for handling a TheOldReader login session."""

    _LOGIN: Final[str] = "https://theoldreader.com/accounts/ClientLogin"
    """The URL of the endpoint used to log in."""

    _API: Final[str] = "https://theoldreader.com/reader/api/0/"
    """The URL of the API endpoint."""

    _USER_AGENT: Final[str] = f"oldas v{__version__} (https://github.com/davep/oldas)"
    """The user agent to use for the library."""

    def __init__(
        self, client: str, auth_code: str | None = None, timeout: int = 60
    ) -> None:
        """Initialise the object.

        Args:
            client: The name of the client that is logging in.
            auth_code: Optional authorisation code to resume a session.
            timeout: The timeout in seconds to use when making calls.

        Note:
            The `client` should be a unique name you give your client
            application that is using this library.
        """
        self._client = client
        """The name of the client to log in as."""
        self._auth_code = auth_code
        """The authorisation code."""
        self._timeout = timeout
        """The timeout, in seconds, to use when making calls."""
        self._web_client_: AsyncClient | None = None
        """The internal reference to the HTTPX client."""

    @property
    def _web_client(self) -> AsyncClient:
        """The HTTPX client."""
        if self._web_client_ is None:
            self._web_client_ = AsyncClient(timeout=self._timeout)
        return self._web_client_

    @property
    def logged_in(self) -> bool:
        """Are we logged in?"""
        return self._auth_code is not None

    @property
    def auth_code(self) -> str | None:
        """The auth code, if we are logged in, else [`None`][None]."""
        return self._auth_code

    async def _call(self, call: Awaitable[Response]) -> Response:
        """Make a call out to the API.

        Args:
            call: The call to make.

        Returns:
            The response.

        Raises:
            OldASError: If there was an error connecting or logging in.
        """
        try:
            response = await call
        except ReadTimeout:
            raise OldASError("Timeout while talking to TheOldReader API")
        except RequestError as error:
            raise OldASError(str(error)) from None
        try:
            response.raise_for_status()
        except HTTPStatusError as error:
            if error.response.status_code == httpx.codes.UNAUTHORIZED:
                raise OldASLoginNeeded("The current token is not valid") from None
            if error.response.status_code == httpx.codes.FORBIDDEN:
                raise OldASInvalidLogin(str(error)) from None
            raise OldASError(str(error)) from None
        return response

    @property
    def _headers(self) -> dict[str, str]:
        """The standard headers for a call to TheOldReader."""
        return {
            "user-agent": self._USER_AGENT,
        } | (
            {"Authorization": f"GoogleLogin auth={self._auth_code}"}
            if self.logged_in
            else {}
        )

    async def login(self, user: str, password: str) -> Self:
        """Log into TheOldReader.

        Args:
            user: The user name to log in with.
            password: The password to log in with.

        Returns:
            Self.

        Raises:
            OldASError: If there was an error connecting or logging in.
        """
        if self._auth_code is None:
            self._auth_code = (
                (
                    await self._call(
                        self._web_client.post(
                            self._LOGIN,
                            json={
                                "accountType": "HOSTED_OR_GOOGLE",
                                "client": self._client,
                                "Email": user,
                                "Passwd": password,
                                "service": "reader",
                                "output": "json",
                            },
                            headers=self._headers,
                        )
                    )
                )
                .json()
                .get("Auth")
            )
        return self

    def logout(self) -> Self:
        """Log out of the TheOldReader.

        Returns:
            Self.
        """
        self._auth_code = None
        return self

    def _must_be_logged_in(self) -> None:
        """Checks if we're logged in and raises an error if not."""
        if not self.logged_in:
            raise OldASLoginNeeded("API call made but not logged in")

    @staticmethod
    def _verify_raw(data: Any) -> RawData:
        """Verify that the given data is of the type we expect.

        Args:
            data: The data that was received.

        Returns:
            The verified data, of the expected type.
        """
        if isinstance(data, dict):
            return data
        raise OldASError("Unexpected data type received from TheOldReader API")

    async def get(self, url: str, **params: Any) -> RawData:
        """Make a GET call to the API.

        Args:
            url: The URL to call.
            params: Any extra parameters that need to be passed.

        Returns:
            A dictionary that is the JSON data.

        Raises:
            OldASError: If there was an error connecting or logging in.
        """
        self._must_be_logged_in()
        return self._verify_raw(
            (
                await self._call(
                    self._web_client.get(
                        f"{self._API}{url}",
                        headers=self._headers,
                        params={**params, "output": "json"},
                    )
                )
            ).json()
        )

    @staticmethod
    def _verify_ok(response: Response) -> bool:
        """Verify that the given data is of the type we expect.

        Args:
            response: The response that was received.

        Returns:
            [`True`][True] if the response reported that everything was
            okay, [`False`][False] if not.
        """
        return response.text.strip() == "OK"

    async def _post(self, url: str, **data: Any) -> Response:
        """Make a POST call to the API.

        Args:
            url: The URL to call.
            data: The data to pass.

        Returns:
            The response from the call to the API.

        Raises:
            OldASError: If there was an error connecting or logging in.
        """
        self._must_be_logged_in()
        return await self._call(
            self._web_client.post(
                f"{self._API}{url}",
                headers=self._headers,
                data=data,
            )
        )

    async def post(self, url: str, **data: Any) -> RawData:
        """Make a POST call to the API.

        Args:
            url: The URL to call.
            data: The data to pass.

        Returns:
            The response from the call to the API.

        Raises:
            OldASError: If there was an error connecting or logging in.
        """
        return self._verify_raw((await self._post(url, **data)).json())

    async def post_ok(self, url: str, **data: Any) -> bool:
        """Make a POST call to the API.

        Args:
            url: The URL to call.
            data: The data to pass.

        Returns:
            [`True`][True] if the call worked, [`False`][False] if not.

        Raises:
            OldASError: If there was an error connecting or logging in.
        """
        return self._verify_ok(await self._post(url, **data))

    async def _edit_tag(
        self, item: str | list[str], tag: str | State, operation: Literal["a", "r"]
    ) -> bool:
        """Perform an edit on a tag.

        Args:
            item: The item(s) to perform the edit on.
            tag: The tag to add or remove.
            operation: The operation to perform.

        Returns:
            [`True`][True] if the tag edit operation worked,
            [`False`][False] if not.
        """
        return await self.post_ok("/edit-tag", i=item, **{str(operation): str(tag)})

    async def add_tag(self, item: str | list[str], tag: str | State) -> bool:
        """Add a tag to an item.

        Args:
            item: The item(s) to perform the edit on.
            tag: The tag to add.

        Returns:
            [`True`][True] if the add tag operation worked,
            [`False`][False] if not.
        """
        return await self._edit_tag(item, tag, "a")

    async def remove_tag(self, item: str | list[str], tag: str | State) -> bool:
        """Remove a tag from an item.

        Args:
            item: The item(s) to perform the edit on.
            tag: The tag to remove.

        Returns:
            [`True`][True] if the remove tag operation worked,
            [`False`][False] if not.
        """
        return await self._edit_tag(item, tag, "r")


### session.py ends here



"""Provides a class for getting and holding the user information."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass

##############################################################################
# Local imports.
from .session import Session
from .types import RawData


##############################################################################
@dataclass
class User:
    """TheOldReader user information."""

    user_id: str
    """The user's ID."""
    name: str
    """The user's name."""
    profile_id: str
    """The user's profile ID."""
    email: str
    """The user's email address."""
    is_blogger_user: str
    """Is the user a Blogger user?"""
    signup_time: int
    """The signup time of the user."""
    is_multi_login_enabled: bool
    """Is multi-login enabled?"""
    is_premium: bool
    """Is the user a premium user?"""

    @classmethod
    async def load(cls, session: Session) -> User:
        """Load the current user's details.

        Args:
            session: The API session to use to load the data.

        Returns:
            The user details.
        """
        user = await session.get("user-info")
        return cls(
            user_id=user["userId"],
            name=user["userName"],
            profile_id=user["userProfileId"],
            email=user["userEmail"],
            is_blogger_user=user["isBloggerUser"],
            signup_time=user["signupTimeSec"],
            is_multi_login_enabled=user["isMultiLoginEnabled"],
            is_premium=user["isPremium"],
        )


### user.py ends here



"""Provides a class for loading up the folders."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass

##############################################################################
# Local imports.
from .prefixes import Prefix, id_is_a_folder
from .session import Session
from .types import OldList, RawData


##############################################################################
@dataclass(frozen=True)
class Folder:
    """Folder information class."""

    id: str
    """The ID of the folder."""
    sort_id: str
    """The sort ID of the folder."""

    @property
    def name(self) -> str:
        """The name of the folder."""
        return self.id.removeprefix(Prefix.FOLDER)

    @classmethod
    def from_json(cls, data: RawData) -> Folder:
        """Load the folder from JSON data.

        Args:
            data: The data to load the folder from.

        Returns:
            The folder information.
        """
        return Folder(
            id=data["id"],
            sort_id=data["sortid"],
        )


##############################################################################
class Folders(OldList[Folder]):
    """Load the [folder][oldas.Folder] list from TheOldReader."""

    @classmethod
    async def load(cls, session: Session) -> Folders:
        """Load the folders.

        Args:
            session: The API session object.

        Returns:
            A list of [folders][oldas.Folder].
        """
        return cls(
            Folder.from_json(folder)
            for folder in (await session.get("tag/list"))["tags"]
            if id_is_a_folder(folder.get("id", ""))
        )

    @staticmethod
    def full_id(folder: str | Folder) -> str:
        """Turn something that identifies a folder into a full folder ID.

        Args:
            folder: The folder to get the ID from.

        Returns:
            The full prefixed folder ID.
        """
        if isinstance(folder, Folder):
            folder = folder.id
        return folder if id_is_a_folder(folder) else f"{Prefix.FOLDER}{folder}"

    @classmethod
    async def rename(
        cls, session: Session, rename_from: str | Folder, rename_to: str
    ) -> bool:
        """Rename a folder on the server.

        Args:
            session: The API session object.
            rename_from: The folder that is to be renamed.
            rename_to: The new name for the folder.

        Returns:
            [`True`][True] if the rename worked, [`False`][False] if not.

        Notes:
            `rename_from` and `rename_to` can have or be missing the prefix
            [`Prefix.FOLDER`][oldas.prefixes.Prefix.FOLDER]; this method
            will handle either case and do the right thing.
        """
        return await session.post_ok(
            "rename-tag", s=cls.full_id(rename_from), dest=cls.full_id(rename_to)
        )

    @classmethod
    async def remove(cls, session: Session, folder: str | Folder) -> bool:
        """Remove the given folder from the server.

        Args:
            session: The API session object.
            folder: The folder that is to be removed.

        Returns:
            [`True`][True] if the folder was removed, [`False`][False] if not.

        Notes:
            `folder` can have or be missing the prefix
            [`Prefix.FOLDER`][oldas.prefixes.Prefix.FOLDER]; this method
            will handle either case and do the right thing.
        """
        return await session.post_ok("disable-tag", s=cls.full_id(folder))


### folders.py ends here



"""Provides a class for getting unread information."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from datetime import datetime, timezone

##############################################################################
# Local imports.
from .prefixes import Prefix, id_is_a
from .session import Session
from .types import OldList, RawData


##############################################################################
@dataclass(frozen=True)
class Count:
    """Unread count information class."""

    id: str
    """The ID of the item that has an unread count."""
    unread: int
    """The unread count."""
    newest_timestamp: datetime
    """The timestamp of the newest item."""
    prefix: str
    """The prefix related to this type of count."""

    @property
    def name(self) -> str:
        """The name of the count."""
        return self.id.removeprefix(self.prefix)

    @classmethod
    def from_json(cls, data: RawData, prefix: str) -> Count:
        """Load the count from JSON data.

        Args:
            data: The data to load the count from.
            prefix: The prefix to associate with this type of count.

        Returns:
            The count information.
        """
        return Count(
            id=data["id"],
            unread=data["count"],
            newest_timestamp=datetime.fromtimestamp(
                int(data["newestItemTimestampUsec"]) / 1_000_000,
                timezone.utc,
            ),
            prefix=prefix,
        )


##############################################################################
class Counts(OldList[Count]):
    """Holds a collection of [counts][oldas.Count]."""


##############################################################################
@dataclass(frozen=True)
class Unread:
    """Class that loads and holds [unread counts][oldas.Counts]."""

    total: int
    """The total unread count."""
    folders: Counts
    """The unread counts for each folder."""
    feeds: Counts
    """The unread count for each feed."""

    @staticmethod
    def _get_counts(unread: RawData, prefixed_with: Prefix) -> Counts:
        """Get a particular set of unread counts.

        Args:
            unread: The unread data.
            prefix_with: The prefix to look for.

        Returns:
            A list of unread counts of the given prefix.
        """
        return Counts(
            Count.from_json(count, prefixed_with)
            for count in unread["unreadcounts"]
            if id_is_a(count["id"], prefixed_with)
        )

    @classmethod
    async def load(cls, session: Session) -> Unread:
        """Load the unread counts.

        Args:
            session: The API session object.

        Returns:
            The unread counts.
        """
        unread = await session.get("unread-count")
        return cls(
            total=unread["max"],
            folders=cls._get_counts(unread, Prefix.FOLDER),
            feeds=cls._get_counts(unread, Prefix.FEED),
        )


### unread.py ends here
